import numpy as np
from qiskit import QuantumCircuit
from qiskit.transpiler import CouplingMap
from sympy.testing.pytest import raises

from stabiliser_suite import clifford, utils


def test___apply_hadamard():
    qc = QuantumCircuit(2)
    tableau = np.array([[0, 0, 1, 0, 1], [1, 1, 0, 1, 0]])
    clifford.__apply_hadamard(tableau, qc, 0)
    expected_tableau = np.array([[1, 0, 0, 0, 1], [0, 1, 1, 1, 0]])
    assert len(qc.data) == 1
    assert qc.data[0].name == 'h'
    assert np.array_equal(tableau, expected_tableau)
    assert qc.data[0].qubits[0]._index == 0

    # Perform the same with shift
    qc = QuantumCircuit(2)
    tableau = np.array([[0, 0, 1, 1, 1], [1, 1, 0, 0, 0]])
    clifford.__apply_hadamard(tableau, qc, 0, shift=1)
    expected_tableau = np.array([[1, 0, 0, 1, 1], [0, 1, 1, 0, 0]])
    assert len(qc.data) == 1
    assert qc.data[0].name == 'h'
    assert np.array_equal(tableau, expected_tableau)
    assert qc.data[0].qubits[0]._index == 1


def test___apply_s():
    qc = QuantumCircuit(2)
    tableau = np.array([[0, 0, 1, 0, 1], [1, 1, 0, 1, 0]])
    clifford.__apply_s(tableau, qc, 0)
    expected_tableau = np.array([[0, 0, 1, 0, 1], [1, 1, 1, 1, 0]])
    assert len(qc.data) == 1
    assert qc.data[0].name == 's'
    assert np.array_equal(tableau, expected_tableau)
    assert qc.data[0].qubits[0]._index == 0

    # Perform the same with shift
    qc = QuantumCircuit(2)
    tableau = np.array([[0, 0, 1, 1, 1], [1, 1, 1, 0, 0]])
    clifford.__apply_s(tableau, qc, 0, shift=1)
    expected_tableau = np.array([[0, 0, 1, 1, 1], [1, 1, 0, 0, 0]])
    assert len(qc.data) == 1
    assert qc.data[0].name == 's'
    assert np.array_equal(tableau, expected_tableau)
    assert qc.data[0].qubits[0]._index == 1


def test___apply_cnot():
    qc = QuantumCircuit(2)
    tableau = np.array([[0, 0, 1, 0, 1], [1, 1, 0, 1, 0]])
    clifford.__apply_cnot(tableau, qc, 0, 1)
    expected_tableau = np.array([[0, 0, 1, 0, 1], [1, 0, 1, 1, 0]])
    assert len(qc.data) == 1
    assert qc.data[0].name == 'cx'
    assert np.array_equal(tableau, expected_tableau)
    assert qc.data[0].qubits[0]._index == 0
    assert qc.data[0].qubits[1]._index == 1

    # Perform the same with shift
    qc = QuantumCircuit(3)
    tableau = np.array([[0, 0, 1, 1, 1, 0, 0], [1, 1, 1, 0, 0, 0, 1]])
    clifford.__apply_cnot(tableau, qc, 0, 1, shift=1)
    expected_tableau = np.array([[0, 0, 1, 0, 1, 0, 0], [1, 0, 1, 0, 0, 0, 1]])
    assert len(qc.data) == 1
    assert qc.data[0].name == 'cx'
    assert np.array_equal(tableau, expected_tableau)
    assert qc.data[0].qubits[0]._index == 1
    assert qc.data[0].qubits[1]._index == 2


def test___apply_swap():
    qc = QuantumCircuit(2)
    tableau = np.array([[0, 0, 1, 0, 1], [1, 1, 0, 1, 0]])
    clifford.__apply_swap(tableau, qc, 0, 1)
    expected_tableau = np.array([[0, 0, 0, 1, 1], [1, 1, 1, 0, 0]])
    assert len(qc.data) == 1
    assert qc.data[0].name == 'swap'
    assert np.array_equal(tableau, expected_tableau)
    assert qc.data[0].qubits[0]._index == 0
    assert qc.data[0].qubits[1]._index == 1

    # Perform the same with shift
    qc = QuantumCircuit(3)
    tableau = np.array([[1, 0, 0, 1, 1, 0, 0], [1, 1, 1, 0, 0, 0, 1]])
    clifford.__apply_swap(tableau, qc, 0, 1, shift=1)
    expected_tableau = np.array([[0, 1, 0, 1, 1, 0, 0], [1, 1, 1, 0, 0, 0, 1]])
    assert len(qc.data) == 1
    assert qc.data[0].name == 'swap'
    assert np.array_equal(tableau, expected_tableau)
    assert qc.data[0].qubits[0]._index == 1
    assert qc.data[0].qubits[1]._index == 2


def test___get_pauli():
    tableau = np.array([[1, 0, 0, 1, 0], [1, 0, 1, 0, 0, ]])
    assert clifford.__get_pauli(tableau, 0, 0) == 'x'
    assert clifford.__get_pauli(tableau, 0, 1) == 'z'
    assert clifford.__get_pauli(tableau, 1, 0) == 'y'
    assert clifford.__get_pauli(tableau, 1, 1) == 'I'


def test___apply_pauli():
    qc = QuantumCircuit(2)
    tableau = np.array([[0, 0, 1, 0, 1], [1, 1, 0, 1, 0]])
    clifford.__apply_pauli(tableau, qc, 0, 'x')
    expected_tableau = np.array([[0, 0, 1, 0, 0], [1, 1, 0, 1, 0]])
    assert len(qc.data) == 1
    assert qc.data[0].name == 'x'
    assert np.array_equal(tableau, expected_tableau)
    assert qc.data[0].qubits[0]._index == 0

    qc = QuantumCircuit(2)
    tableau = np.array([[0, 0, 1, 0, 1], [1, 1, 0, 1, 0]])
    clifford.__apply_pauli(tableau, qc, 0, 'y', shift=1)
    expected_tableau = np.array([[0, 0, 1, 0, 0], [1, 1, 0, 1, 1]])
    assert len(qc.data) == 1
    assert qc.data[0].name == 'y'
    assert np.array_equal(tableau, expected_tableau)
    assert qc.data[0].qubits[0]._index == 1

    qc = QuantumCircuit(2)
    tableau = np.array([[0, 0, 1, 0, 1], [1, 1, 0, 1, 0]])
    clifford.__apply_pauli(tableau, qc, 0, 'z', shift=1)
    expected_tableau = np.array([[0, 0, 1, 0, 1], [1, 1, 0, 1, 1]])
    assert len(qc.data) == 1
    assert qc.data[0].name == 'z'
    assert np.array_equal(tableau, expected_tableau)
    assert qc.data[0].qubits[0]._index == 1

    with raises(ValueError):
        clifford.__apply_pauli(tableau, qc, 0, 'a', shift=1)


def test_print_tableau(capsys):
    tableau = np.array([[0, 0, 1, 0, 1], [1, 1, 0, 1, 0]])
    clifford.print_tableau(tableau, representation="text")
    captured = capsys.readouterr()
    expected = ("┌────┬────┬────┬────┬───┐\n│ X0 │ X1 │ Z0 │ Z1 │ S │\n╞════╪════╪════╪════╪═══╡\n│ 0  │ 0  │ 1  │ 0  "
                "│ 1 │\n├────┼────┼────┼────┼───┤\n│ 1  │ 1  │ 0  │ 1  │ 0 │\n└────┴────┴────┴────┴───┘\n")
    assert captured.out == expected

    with raises(NotImplementedError):
        clifford.print_tableau(tableau, representation="mpl")

    with raises(ValueError):
        clifford.print_tableau(tableau, representation="aaa")


def test___check_normalisation():
    tableau = np.array([[0, 0, 1, 0, 1], [1, 1, 0, 1, 0]])
    assert clifford.__check_normalisation(tableau) == False
    tableau = np.array([[1, 0, 0, 0, 0], [0, 0, 1, 0, 0]])
    assert clifford.__check_normalisation(tableau) == True


def test___step_1():
    tableau = np.array([[1, 0, 0, 1, 1, 0, 0], [1, 1, 0, 1, 1, 0, 0]])
    qc = QuantumCircuit(3)
    clifford.__step_1(tableau, qc)
    assert len(qc.data) == 2
    assert qc.data[0].name == 's'
    assert qc.data[0].qubits[0]._index == 0
    assert qc.data[1].name == 'h'
    assert qc.data[1].qubits[0]._index == 1
    expected_tableau = np.array([[1, 1, 0, 0, 0, 0, 0], [1, 1, 0, 0, 1, 0, 0]])
    assert np.array_equal(tableau, expected_tableau)


def test___step_2():
    qc = QuantumCircuit(4)
    tableau = np.array([[1, 1, 1, 1, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 0, 0, 0, 0]])
    clifford.__step_2(tableau, qc)
    assert len(qc.data) == 3
    assert qc.data[0].name == 'cx'
    assert qc.data[0].qubits[0]._index == 0
    assert qc.data[0].qubits[1]._index == 1
    assert qc.data[1].name == 'cx'
    assert qc.data[1].qubits[0]._index == 2
    assert qc.data[1].qubits[1]._index == 3
    assert qc.data[2].name == 'cx'
    assert qc.data[2].qubits[0]._index == 0
    assert qc.data[2].qubits[1]._index == 2

    expected_tableau = np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0], [1, 0, 0, 0, 1, 0, 0, 0, 0]])
    assert np.array_equal(tableau, expected_tableau)


def test___step_3():
    qc = QuantumCircuit(3)
    tableau = np.array([[0, 1, 0, 0, 0, 0, 0], [1, 1, 0, 1, 1, 0, 0]])
    clifford.__step_3(tableau, qc)
    assert len(qc.data) == 1
    assert qc.data[0].name == 'swap'
    assert qc.data[0].qubits[0]._index == 0
    assert qc.data[0].qubits[1]._index == 1
    expected_tableau = tableau = np.array([[1, 0, 0, 0, 0, 0, 0], [1, 1, 0, 1, 1, 0, 0]])
    assert np.array_equal(tableau, expected_tableau)

    with raises(RuntimeError):
        qc = QuantumCircuit(3)
        tableau = np.array([[1, 1, 1, 0, 0, 0, 0], [1, 1, 0, 1, 1, 0, 0]])
        clifford.__step_3(tableau, qc)


def test___step_4():
    qc = QuantumCircuit(3)
    tableau = np.array([[1, 0, 0, 0, 0, 0, 0], [1, 1, 0, 1, 1, 0, 0]])
    clifford.__step_4(tableau, qc)
    assert len(qc.data) == 1
    assert qc.data[0].name == 'h'
    assert qc.data[0].qubits[0]._index == 0
    expected_tableau = np.array([[0, 0, 0, 1, 0, 0, 0], [1, 1, 0, 1, 1, 0, 0]])
    assert np.array_equal(tableau, expected_tableau)

    qc = QuantumCircuit(2)
    tableau = np.array([[1, 0, 0, 0, 0], [0, 1, 1, 0, 0]])
    clifford.__step_4(tableau, qc)
    assert len(qc.data) == 0
    expected_tableau = np.array([[1, 0, 0, 0, 0], [0, 1, 1, 0, 0]])
    assert np.array_equal(tableau, expected_tableau)


def test___step_5():
    qc = QuantumCircuit(1)
    tableau = np.array([[1, 0, 0], [0, 1, 1]])
    clifford.__step_5(tableau, qc)
    assert len(qc.data) == 1
    assert qc.data[0].name == 'x'
    assert qc.data[0].qubits[0]._index == 0
    expected_tableau = np.array([[1, 0, 0], [0, 1, 0]])
    assert np.array_equal(tableau, expected_tableau)

    qc = QuantumCircuit(1)
    tableau = np.array([[1, 0, 1], [0, 1, 1]])
    clifford.__step_5(tableau, qc)
    assert len(qc.data) == 1
    assert qc.data[0].name == 'y'
    assert qc.data[0].qubits[0]._index == 0
    expected_tableau = np.array([[1, 0, 0], [0, 1, 0]])
    assert np.array_equal(tableau, expected_tableau)

    qc = QuantumCircuit(1)
    tableau = np.array([[1, 0, 1], [0, 1, 0]])
    clifford.__step_5(tableau, qc)
    assert len(qc.data) == 1
    assert qc.data[0].name == 'z'
    assert qc.data[0].qubits[0]._index == 0
    expected_tableau = np.array([[1, 0, 0], [0, 1, 0]])
    assert np.array_equal(tableau, expected_tableau)


def test___generate_random_tableau():
    tab1 = clifford.__generate_random_tableau(8)
    tab2 = clifford.__generate_random_tableau(8)
    assert not np.array_equal(tab1, tab2)


def test___perform_sweeping():
    qc = QuantumCircuit(4)
    tableau = np.array([[1, 1, 1, 1, 0, 1, 1, 0, 0], [1, 1, 1, 1, 1, 1, 1, 0, 1]])
    clifford.__perform_sweeping(tableau, qc)
    assert len(qc.data) == 9
    assert qc.data[0].name == 's'
    assert qc.data[0].qubits[0]._index == 1
    assert qc.data[1].name == 's'
    assert qc.data[1].qubits[0]._index == 2
    assert qc.data[2].name == 'cx'
    assert qc.data[2].qubits[0]._index == 0
    assert qc.data[2].qubits[1]._index == 1
    assert qc.data[3].name == 'cx'
    assert qc.data[3].qubits[0]._index == 2
    assert qc.data[3].qubits[1]._index == 3
    assert qc.data[4].name == 'cx'
    assert qc.data[4].qubits[0]._index == 0
    assert qc.data[4].qubits[1]._index == 2
    assert qc.data[5].name == 'h'
    assert qc.data[5].qubits[0]._index == 0
    assert qc.data[6].name == 's'
    assert qc.data[6].qubits[0]._index == 0
    assert qc.data[7].name == 'h'
    assert qc.data[7].qubits[0]._index == 0
    assert qc.data[8].name == 'x'
    assert qc.data[8].qubits[0]._index == 0

    expected_tableau = np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 1, 0, 0, 0, 0]])
    assert np.array_equal(tableau, expected_tableau)

    qc = QuantumCircuit(3)
    tableau = np.array([[0, 0, 0, 0, 1, 0, 0], [1, 1, 0, 1, 1, 0, 0]])
    clifford.__perform_sweeping(tableau, qc)
    assert len(qc.data) == 7
    assert qc.data[0].name == 'h'
    assert qc.data[0].qubits[0]._index == 1
    assert qc.data[1].name == 'swap'
    assert qc.data[1].qubits[0]._index == 0
    assert qc.data[1].qubits[1]._index == 1
    assert qc.data[2].name == 'h'
    assert qc.data[2].qubits[0]._index == 0
    assert qc.data[3].name == 's'
    assert qc.data[3].qubits[0]._index == 0
    assert qc.data[4].name == 's'
    assert qc.data[4].qubits[0]._index == 1
    assert qc.data[5].name == 'cx'
    assert qc.data[5].qubits[0]._index == 0
    assert qc.data[5].qubits[1]._index == 1
    assert qc.data[6].name == 'h'
    assert qc.data[6].qubits[0]._index == 0
    expected_tableau = np.array([[1, 0, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0]])
    assert np.array_equal(tableau, expected_tableau)

    qc = QuantumCircuit(2)
    tableau = np.array([[0, 1, 0, 0, 0], [0, 0, 0, 1, 0]])
    clifford.__perform_sweeping(tableau, qc)
    assert len(qc.data) == 1
    assert qc.data[0].name == 'swap'
    assert qc.data[0].qubits[0]._index == 0
    assert qc.data[0].qubits[1]._index == 1
    expected_tableau = np.array([[1, 0, 0, 0, 0], [0, 0, 1, 0, 0]])
    assert np.array_equal(tableau, expected_tableau)

    qc = QuantumCircuit(1)
    tableau = np.array([[0, 1, 0], [1, 0, 0]])
    clifford.__perform_sweeping(tableau, qc)
    assert len(qc.data) == 1
    assert qc.data[0].name == 'h'
    assert qc.data[0].qubits[0]._index == 0
    expected_tableau = np.array([[1, 0, 0], [0, 1, 0]])
    assert np.array_equal(tableau, expected_tableau)


def test___check_if_tableau_suits():
    tableau = np.array([[1, 0, 0, 0, 1, 0, 0], [1, 0, 1, 1, 0, 0, 0]])
    assert clifford.__check_if_tableau_suits(tableau) is True

    tableau = np.array([[0, 0, 0], [0, 1, 0]])
    assert clifford.__check_if_tableau_suits(tableau) is False


def test___generate_anticommuting_tablueau():
    row = clifford.__generate_anticommuting_tableau(3)
    assert clifford.__check_if_tableau_suits(row) is True
    assert utils.__find_qubit_count(row) == 3
    assert utils.__do_paulis_anticommute(row) is True

def test___step_3_with_coupling_map():
    qc: QuantumCircuit = QuantumCircuit(2)
    row: np.ndarray = np.array([[0,1,0,0,0], [0,1,0,0,0]])

    coupling_map = CouplingMap([])
    assert clifford.__step_3(row, qc, coupling_map=coupling_map, max_distance=1) is False
    coupling_map = CouplingMap([[0, 2], [1 ,2]])
    assert clifford.__step_3(row, qc, coupling_map=coupling_map, max_distance=1) is False
    assert clifford.__step_3(row, qc, coupling_map=coupling_map, max_distance=2) is True

def test___step_2_with_coupling_map():
    qc: QuantumCircuit = QuantumCircuit(3)
    row: np.ndarray = np.array([[1, 1, 1, 0, 0, 0, 0], [1, 0, 0, 1, 0, 0, 0]])

    coupling_map = CouplingMap([])
    assert clifford.__step_2(row, qc, coupling_map=coupling_map, max_distance=1) is False
    coupling_map = CouplingMap([[0, 2], [1, 2]])
    assert clifford.__step_2(row, qc, coupling_map=coupling_map, max_distance=1) is False
    assert clifford.__step_2(row, qc, coupling_map=coupling_map, max_distance=2) is True


def test___sample_clifford_group():
    coupling_map = CouplingMap([[0, 1], [1, 2]])
    qc = clifford.sample_clifford_group(3, True, coupling_map=coupling_map, max_distance=1)
    assert len(qc.qubits) == 3
    assert len(qc.data) > 0



from __future__ import annotations
import inspect
import itertools
import math
from datetime import datetime, timezone
from typing import Tuple, Sequence, Dict, List

import numpy as np
from numba import njit
from numba.core.types import complex128
from numpy import ndarray
from qiskit import QiskitError
from qiskit.circuit import Gate, Instruction, CircuitError
from qiskit.circuit.library import iSwapGate, XGate, YGate, ZGate, IGate, HGate, SGate, CZGate, SwapGate, \
    standard_gates, CXGate
from qiskit.quantum_info import Clifford
from tqdm import tqdm

# Define basic gates
I = np.array([[1, 0], [0, 1]], dtype=complex)
X = np.array([[0, 1], [1, 0]], dtype=complex)
Y = np.array([[0, -1j], [1j, 0]], dtype=complex)
Z = np.array([[1, 0], [0, -1]], dtype=complex)
H = (1 / np.sqrt(2)) * np.array([[1, 1], [1, -1]], dtype=complex)
S = np.array([[1, 0], [0, 1j]], dtype=complex)


@njit(cache=True)
def rx(theta):
    """
        Generate the rotation matrix for the X-axis (Rx) in quantum computing.

        This function creates a 2x2 unitary matrix representing a rotation
        around the X-axis by a given angle `theta`.

        Args:
            theta (float): The rotation angle in radians.

        Returns:
            np.ndarray: A 2x2 complex-valued numpy array representing the Rx matrix.
        """
    return np.array([
        [np.cos(theta / 2), -1j * np.sin(theta / 2)],
        [-1j * np.sin(theta / 2), np.cos(theta / 2)]
    ], dtype=complex128)


@njit(cache=True)
def rz(theta):
    """
    Generate the rotation matrix for the Z-axis (Rz) in quantum computing.

    This function creates a 2x2 unitary matrix representing a rotation
    around the Z-axis by a given angle `theta`.

    Args:
        theta (float): The rotation angle in radians.

    Returns:
        np.ndarray: A 2x2 complex-valued numpy array representing the Rz matrix.
    """
    return np.array([
        [np.exp(-1j * theta / 2), 0],
        [0, np.exp(1j * theta / 2)]
    ], dtype=complex128)


def __generate_two_qubit_gates(gates_dict):
    tensor_gates = {}
    for name1, gate1 in gates_dict.items():
        for name2, gate2 in gates_dict.items():
            name = f"{name1}⊗{name2}"
            tensor_gates[name] = np.kron(gate1, gate2)

    # Add special two-qubit gates
    tensor_gates.update({
        CXGate().name: CXGate().to_matrix(),
        iSwapGate().name: iSwapGate().to_matrix(),
        CZGate().name: CZGate().to_matrix(),
        SwapGate().name: SGate().to_matrix()
    })

    return tensor_gates


def generate_two_qubit_clifford_gates():
    """
    Generate a dictionary of two-qubit Clifford gates.

    This function creates a collection of two-qubit Clifford gates, including
    standard single-qubit gates, rotation gates, and special gates like iSwap and CZ.
    It also generates tensor products of sinqle-qubit gates to form two-qubit gates.

    Returns:
        dict: A dictionary where keys are gate names (str) and values are 2D numpy arrays
              representing the corresponding gate matrices.
    """
    # Define standard single-qubit gates
    gates = {
        IGate().name: IGate().to_matrix(),
        XGate().name: XGate().to_matrix(),
        YGate().name: YGate().to_matrix(),
        ZGate().name: ZGate().to_matrix(),
        HGate().name: HGate().to_matrix(),
        SGate().name: SGate().to_matrix(),
    }

    # Define rotation gates with various angles
    thetas = {
        'pi/2': np.pi / 2,
        'pi': np.pi, '3pi/2': 3 * np.pi / 2, '-3pi/2': -3 * np.pi / 2, '-pi': -np.pi, '-pi/2': - np.pi / 2
    }
    for theta, angle in thetas.items():
        gates.update({
            f'Rx({theta})': rx(angle),
            f'Rz({theta})': rz(angle)
        })

    # Generate tensor products of the gates
    return __generate_two_qubit_gates(gates)


@njit
def __are_equal_up_to_global_phase(u1, u2, atol=1e-8):
    """
        Check if two unitary matrices are equal up to a global phase.

        This function determines whether two unitary matrices, `U1` and `U2`,
        are equivalent up to a global phase factor. A global phase factor means
        that the matrices differ only by a scalar multiplication of a complex
        exponential (e.g., e^(i*theta)).

        Args:
            u1 (np.ndarray): The first unitary matrix.
            u2 (np.ndarray): The second unitary matrix.
            atol (float): Absolute tolerance for numerical comparison. Default is 1e-8.

        Returns:
            bool: True if the matrices are equal up to a global phase, False otherwise.
        """
    u = u1 @ u2.conj().T
    phase = np.diag(u)[0]

    return np.allclose(u, phase * np.eye(u.shape[0]), atol=atol)


@njit
def __multiply_sequence(matrices):
    """
    Multiply a sequence of matrices in order.

    This function takes a list of matrices and computes their product
    in the given order. The multiplication is performed iteratively.

    Args:
        matrices (list of np.ndarray): A list of 2D numpy arrays (matrices) to be multiplied.

    Returns:
        np.ndarray: The resulting matrix after multiplying all matrices in the sequence.
    """
    result = np.eye(matrices[0].shape[0], dtype=complex128)
    for m in matrices:
        result = result @ m
    return result


def find_matching_combinations(gate_dict, target_matrix, output=None, max_depth=3, allow_global_phase=True) -> list[
    tuple[str, np.ndarray]]:
    results = []
    try:
        gate_names = list(gate_dict.keys())

        for depth in range(1, max_depth + 1):
            print(f"Checking depth {depth} ({len(gate_names) ** depth} combinations)...")
            for sequence in tqdm(itertools.product(gate_names, repeat=depth), total=len(gate_names) ** depth):
                result = np.eye(target_matrix.shape[0], dtype=complex)
                for gate_name in sequence:
                    if result.shape == gate_dict[gate_name].shape:
                        result = result @ gate_dict[gate_name]
                if (allow_global_phase and __are_equal_up_to_global_phase(result, target_matrix)) or \
                        (not allow_global_phase and np.allclose(result, target_matrix, atol=1e-8)):
                    results.append((sequence, result))

            solutions = [a for a, _ in results]
            print(f"Current solutions at depth {depth}: {solutions}", flush=True)
    finally:
        if output is not None:
            with open(output, 'a', encoding='utf-8') as f:
                now = datetime.now(timezone.utc).astimezone()
                now_text = now.strftime('%Y-%m-%d %H:%M:%S %Z%z')
                f.write(f'\n{now_text} - {max_depth} depth search\n')
                if len(results) == 0:
                    f.write('NO SOLUTIONS FOUND\n')
                    f.flush()
                else:
                    f.write('COMPLETE RESULTS:')
                    f.write(', '.join(f'({a}, {b})' for a, b in results))
                    f.write('\n')
                    f.flush()
    return results


def is_clifford_gate(gate: Gate) -> bool:
    """
    Check if a given quantum gate is a Clifford gate.

    This function determines whether a provided quantum gate belongs to the
    Clifford group. The Clifford group consists of gates that map Pauli operators
    to other Pauli operators under conjugation.

    Args:
        gate (Gate): A quantum gate object from Qiskit.

    Returns:
        bool: True if the gate is a Clifford gate, False otherwise.
    """
    try:
        Clifford(gate)
        return True
    except QiskitError:
        return False


def get_gates(control: int = 0, target: int = 0, num_control_qubits: int = 1) -> dict[str, Gate | Instruction]:
    """
    Retrieve a dictionary of quantum gates with optional control and target qubits.

    This function iterates through the standard gates available in Qiskit,
    attempts to instantiate them, and generates parameterized versions of
    gates if applicable. The resulting dictionary includes gate names as keys
    and their corresponding gate instances as values.

    Args:
        control (int, optional): The control qubit index for gates requiring control qubits. Defaults to 0.
        target (int, optional): The target qubit index for gates requiring target qubits. Defaults to 0.
        num_control_qubits (int, optional): The number of control qubits for gates requiring this parameter. Defaults to 1.

    Returns:
        dict[str, Gate | Instruction]: A dictionary where keys are gate names (optionally parameterized)
                                       and values are the corresponding gate instances.
    """
    gate_classes = {}
    for name, gate_class in vars(standard_gates).items():
        # Check if the item is a class and a subclass of Gate
        if inspect.isclass(gate_class) and issubclass(gate_class, Gate):
            # Attempt to instantiate the gate class
            gates = __try_instantiate_gate(gate_class, control=control, target=target,
                                           num_control_qubits=num_control_qubits)
            if gates:
                # Generate parameterized gate names with angles in steps of π/4
                thetas = [f"{i}pi/4" if i != 0 else "0" for i in range(-6, 7)]
                gate_classes.update(
                    # Add parameterized gates if multiple instances are generated
                    {f"{name}({theta})": gate for theta, gate in zip(thetas, gates)}
                    if len(gates) > 1 else {name: gates[0]}  # Add single instance if only one gate is generated
                )
    return gate_classes


def get_non_clifford_gates(control: int = 0, target: int = 1, num_control_qubits: int = 1) -> dict:
    """
    Retrieve a dictionary of non-Clifford quantum gates.

    This function filters out Clifford gates from the set of all available gates
    and returns only the non-Clifford gates. It uses the `get_gates` function to
    generate all gates and the `is_clifford_gate` function to identify Clifford gates.

    Args:
        control (int, optional): The control qubit index for gates requiring control qubits. Defaults to 0.
        target (int, optional): The target qubit index for gates requiring target qubits. Defaults to 0.
        num_control_qubits (int, optional): The number of control qubits for gates requiring this parameter. Defaults to 1.

    Returns:
        dict: A dictionary where keys are gate names and values are the corresponding
              non-Clifford gate instances.
    """
    all_gate_classes = get_gates(control=control, target=target, num_control_qubits=num_control_qubits)
    gates = {name: gate_class for name, gate_class in all_gate_classes.items() if not is_clifford_gate(gate_class)}
    return gates


def get_clifford_gates(control: int = 0, target: int = 1, num_control_qubits: int = 1) -> dict:
    """
    Retrieve a dictionary of non-Clifford quantum gates.

    This function filters out non-Clifford gates from the set of all available gates
    and returns only the Clifford gates. It uses the `get_gates` function to
    generate all gates and the `is_clifford_gate` function to identify Clifford gates.

    Args:
        control (int, optional): The control qubit index for gates requiring control qubits. Defaults to 0.
        target (int, optional): The target qubit index for gates requiring target qubits. Defaults to 0.
        num_control_qubits (int, optional): The number of control qubits for gates requiring this parameter. Defaults to 1.

    Returns:
        dict: A dictionary where keys are gate names and values are the corresponding
              Clifford gate instances.
    """
    all_gate_classes = get_gates(control=control, target=target, num_control_qubits=num_control_qubits)
    gates = {name: gate_class for name, gate_class in all_gate_classes.items() if is_clifford_gate(gate_class)}
    return gates


def find_non_clifford_transformations(gate: Gate, max_depth: int = 3,
                                      allow_global_phase: bool = True) -> list[tuple[Sequence[str], ndarray]]:
    substitutions = []
    matrices = {}
    if gate.num_qubits == 1:
        all_gates = get_non_clifford_gates()
        matrices = {name: gate.to_matrix() for name, gate in all_gates.items() if gate.num_qubits == 1}
    else:
        all_gates = get_non_clifford_gates()
        matrices = __generate_multi_qubit_gates(all_gates, gate.num_qubits)
    substitutions = mitm_find_transformations(matrices, gate.to_matrix(), max_depth=max_depth,
                                              allow_global_phase=allow_global_phase)
    return substitutions


def find_clifford_transformations(gate: Gate, max_depth: int = 3,
                                  allow_global_phase: bool = True) -> list[tuple[Sequence[str], ndarray]]:
    substitutions = []
    matrices = {}
    if gate.num_qubits == 1:
        all_gates = get_clifford_gates()
        matrices = {name: gate.to_matrix() for name, gate in all_gates.items() if gate.num_qubits == 1}
    else:
        all_gates = get_clifford_gates()
        matrices = __generate_multi_qubit_gates(all_gates, gate.num_qubits)
    substitutions = mitm_find_transformations(matrices, gate.to_matrix(), max_depth=max_depth,
                                              allow_global_phase=allow_global_phase)
    return substitutions


def __try_instantiate_gate(gate_class, control: int = 0, target: int = 1, num_control_qubits: int = 1) -> list[Gate]:
    """
    Attempt to instantiate a quantum gate class or generate parameterized instances.

    This function tries to create an instance of the provided quantum gate class.
    If the class is not a subclass of `Gate`, it returns an empty list. If the class
    requires parameters for instantiation, it generates instances based on the
    required parameters:
    - For gates requiring `num_ctrl_qubits`, it creates an instance with the specified
      number of control qubits.
    - For gates requiring a single parameter `phi`, it generates instances with angles
      ranging from -3π/2 to 3π/2 in steps of π/4.
    - For gates requiring two parameters, it uses the provided `control` and `target` values.

    Args:
        gate_class (type): The class of the quantum gate to instantiate.
        control (int, optional): The control qubit index for two-parameter gates. Defaults to 0.
        target (int, optional): The target qubit index for two-parameter gates. Defaults to 1.
        num_control_qubits (int, optional): The number of control qubits for gates requiring this parameter. Defaults to 1.

    Returns:
        list[Gate]: A list of instantiated quantum gate objects, or an empty list
                    if the class is not a valid quantum gate class or cannot be instantiated.
    """
    if not issubclass(gate_class, Gate):
        return []
    try:
        return [gate_class()]
    except TypeError:

        sig = inspect.signature(gate_class.__init__)
        # Filter out 'self' and kwargs
        required_params = [
            p for p in sig.parameters.values()
            if p.name != 'self' and p.default == inspect.Parameter.empty and p.kind in (
                inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
        ]

        # Generate phis in range (-3π/2, 3π/2) with step π/4
        phis = [i * np.pi / 4 for i in range(-6, 7, 1)]

        try:
            if len(required_params) == 1 and required_params[0].name == 'num_ctrl_qubits':
                return [gate_class(num_control_qubits)]
            elif len(required_params) == 1 and required_params[0].name == 'phi':
                return [gate_class(phi) for phi in phis]
            elif len(required_params) == 2:
                return [gate_class(control, target)]
        except (TypeError, CircuitError):
            return []

    return []


def __generate_multi_qubit_gates_old(gates_dict: dict[str, Gate], target_qubits: int = 2) -> (
        dict)[str, np.ndarray]:
    """
    Generate multi-qubit gate matrices from a dictionary of gates.

    This function creates a dictionary of multi-qubit gates by combining gates using the Kronecker product. It
    ensures that the resulting gates match the specified target qubit dimension.

    Args:
        gates_dict (dict[str, Gate]): A dictionary where keys are gate names and values
                                      are gate objects.
        target_qubits (int, optional): The number of qubits for the target gates. Defaults to 2.

    Returns:
        dict[str, np.ndarray]: A dictionary where keys are gate names (or combinations of gate names)
                               and values are the corresponding multi-qubit gate matrices.
    """
    multi_qubit_gates = {}
    dim = target_qubits ** 2

    gates = gates_dict.copy()
    gates[IGate().name] = IGate().to_matrix()

    for name1, gate1 in gates.items():
        matrix1 = gate1.to_matrix()
        if matrix1.shape[0] == dim:
            multi_qubit_gates[name1] = matrix1
            continue

        for name2, gate2 in gates.items():
            matrix2 = gate2.to_matrix()
            # Check if the combined dimensions match the target dimension
            if matrix1.shape[0] * matrix2.shape[0] == dim:
                multi_qubit_gates[f"{name1}⊗{name2}"] = np.kron(matrix1, matrix2)

    return multi_qubit_gates


def __generate_multi_qubit_gates(gates_dict: dict[str, Gate],
                                 target_qubits: int = 2) -> dict[str, np.ndarray]:
    """
    Generate arbitrary multi-qubit gates (up to target_qubits) by taking
    tensor products of the supplied gates.

    Parameters
    ----------
    gates_dict : dict[str, Gate]
        Dictionary mapping gate names to Qiskit Gate objects.  The routine
        assumes each gate's ``num_qubits`` attribute is correct.
    target_qubits : int
        The desired width (number of qubits) of every generated matrix.

    Returns
    -------
    dict[str, np.ndarray]
        Keys are Kronecker-product strings such as "H⊗X⊗S", values are the
        corresponding unitary matrices of dimension ``2**target_qubits``.
    """
    # ————————————————————————————————————————————————————————————————
    # 1)  Build a catalogue of (name, matrix, arity) including the identity
    # ————————————————————————————————————————————————————————————————
    gate_specs: list[tuple[str, np.ndarray, int]] = []

    for name, gate in gates_dict.items():
        mat = gate.to_matrix()
        gate_specs.append((name, mat, gate.num_qubits))

    # ensure the identity is present for padding
    i_name, i_mat = IGate().name, IGate().to_matrix()
    gate_specs.append((i_name, i_mat, 1))

    # Map arity → list of specs of that size (helps pruning)
    by_arity: dict[int, list[tuple[str, np.ndarray, int]]] = {}
    for spec in gate_specs:
        by_arity.setdefault(spec[2], []).append(spec)

    # ————————————————————————————————————————————————————————————————
    # 2)  DFS to build all tensor products that sum to target_qubits
    # ————————————————————————————————————————————————————————————————
    multi_qubit_gates: dict[str, np.ndarray] = {}

    def dfs(current_name: str,
            current_mat: np.ndarray | None,
            qubits_left: int) -> None:
        """Recursively build Kronecker products totalling target_qubits."""
        if qubits_left == 0:
            multi_qubit_gates[current_name] = current_mat
            return

        # iterate over every gate whose arity ≤ qubits_left
        for arity in range(1, qubits_left + 1):
            for name, mat, _ in by_arity.get(arity, []):
                new_name = name if current_name == "" else f"{current_name}⊗{name}"
                new_mat = mat if current_mat is None else np.kron(current_mat, mat)
                dfs(new_name, new_mat, qubits_left - arity)

    dfs("", None, target_qubits)
    return multi_qubit_gates


def _canonicalise(mat: np.ndarray, *, allow_phase: bool) -> np.ndarray:
    """Return matrix divided by global phase (if allowed)."""
    if not allow_phase:
        return mat
    # find first non-zero entry
    idx = np.flatnonzero(mat)
    if idx.size == 0:
        return mat
    r, c = divmod(idx[0], mat.shape[0])
    phase = mat[r, c] / abs(mat[r, c])
    return mat / phase


def _hash_key(mat: np.ndarray, *, allow_phase: bool) -> Tuple[float, ...]:
    """Convert (canonical) matrix to a hashable tuple with rounding."""
    canon = _canonicalise(mat, allow_phase=allow_phase)
    rounded = np.round(canon.real, 12) + 1j * np.round(canon.imag, 12)
    return tuple(rounded.ravel())


def _matrix_equal(a: np.ndarray, b: np.ndarray,
                  *, allow_phase: bool) -> bool:
    """Return True if a == e^{iθ} b (or exactly equal if phase not allowed)."""
    if allow_phase:
        return np.allclose(_canonicalise(a, allow_phase=True),
                           _canonicalise(b, allow_phase=True),
                           atol=1e-8)
    return np.allclose(a, b, atol=1e-8)


def _enumerate_layers(gate_dict: Dict[str, np.ndarray],
                      depth: int,
                      allow_phase: bool
                      ) -> Dict[Tuple[float, ...], Tuple[Sequence[str], np.ndarray]]:
    """
    Enumerate all products of exactly `depth` gates.

    Returns a dict {hash_key → (sequence, matrix)} keeping the lexicographically
    smallest sequence per hash key.
    """
    if depth == 0:
        ident = np.eye(next(iter(gate_dict.values())).shape[0], dtype=complex)
        key = _hash_key(ident, allow_phase=allow_phase)
        return {key: (tuple(), ident)}

    layer: Dict[Tuple[float, ...], Tuple[Sequence[str], np.ndarray]] = {}
    gate_items = list(gate_dict.items())

    # start from depth-1 layer
    prev = _enumerate_layers(gate_dict, depth - 1, allow_phase)

    for seq, mat in prev.values():
        for gname, gmat in gate_items:
            new_seq = seq + (gname,)
            new_mat = gmat @ mat  # right-multiplication
            key = _hash_key(new_mat, allow_phase=allow_phase)
            # keep lexicographically minimal sequence for determinism
            if key not in layer or new_seq < layer[key][0]:
                layer[key] = (new_seq, new_mat)
    return layer


def mitm_find_transformations(gate_dict: Dict[str, np.ndarray],
                              target_matrix: np.ndarray,
                              *,
                              max_depth: int = 3,
                              allow_global_phase: bool = True
                              ) -> List[Tuple[Sequence[str], np.ndarray]]:
    """
    Meet-in-the-middle search for gate sequences whose product equals
    `target_matrix` (up to optional global phase).

    Parameters
    ----------
    gate_dict
        Mapping {gate_name → unitary matrix}.
    target_matrix
        Target unitary.
    max_depth
        Maximum total length of a candidate sequence.
    allow_global_phase
        If True the match is checked up to a global phase.

    Returns
    -------
    List of (sequence, product_matrix) for every solution found.
    """
    results: List[Tuple[Sequence[str], np.ndarray]] = []
    dim = target_matrix.shape[0]
    ident = np.eye(dim, dtype=complex)
    gates = gate_dict.copy()
    gates[IGate().name] = ident

    # split depth (ceil/ﬂoor)
    left_depth = math.ceil(max_depth / 2)
    right_depth = max_depth - left_depth

    # pre-compute all left products up to left_depth
    left_tables = []
    for d in range(left_depth + 1):
        left_tables.append(_enumerate_layers(gates, d,
                                             allow_phase=allow_global_phase))

    # enumerate right side from depth 0 … right_depth
    for d_r in range(right_depth + 1):
        right_table = _enumerate_layers(gates, d_r,
                                        allow_phase=allow_global_phase)

        for r_key, (r_seq, r_mat) in right_table.items():
            # Needed left matrix to satisfy L · R = U_target
            needed = target_matrix @ np.linalg.inv(r_mat)
            needed_key = _hash_key(needed, allow_phase=allow_global_phase)

            # check every possible left depth up to remaining budget
            max_left_index = min(left_depth, max_depth - d_r)
            for d_l in range(max_left_index, -1, -1):
                if needed_key in left_tables[d_l]:
                    l_seq, l_mat = left_tables[d_l][needed_key]
                    # sanity check
                    prod = r_mat @ l_mat
                    if _matrix_equal(prod, target_matrix,
                                     allow_phase=allow_global_phase):
                        results.append((l_seq + r_seq, prod))

    return results



from stabiliser_suite import utils

import numpy as np


def test__find_qubit_count() -> None:
    tableau = np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]])
    assert utils.__find_qubit_count(tableau) == 2


def test__do_paulis_anticommute() -> None:
    tableau = np.array([[1, 0, 0, 0, 1, 0, 0], [1, 0, 1, 1, 0, 0, 0]])
    assert utils.__do_paulis_anticommute(tableau) == True



import numpy as np

from qiskit.circuit import Instruction
from qiskit.circuit.library import iSwapGate, HGate, TGate, RZGate, XGate, CXGate, MCXGate, CZGate, SwapGate
from qiskit.circuit.gate import Gate
from sympy.testing.pytest import raises

from stabiliser_suite.gate_utils import __are_equal_up_to_global_phase, rx, rz, __generate_two_qubit_gates, \
    generate_two_qubit_clifford_gates, __multiply_sequence, is_clifford_gate, \
    get_non_clifford_gates, __try_instantiate_gate, get_gates, find_non_clifford_transformations


def test___are_equal_up_to_global_phase():
    matrix1 = np.array([[1, 0], [0, 1]], dtype=complex)
    matrix2 = np.array([[1, 0], [0, 1]], dtype=complex)
    assert __are_equal_up_to_global_phase(matrix1, matrix2)

    matrix1 = np.array([[1, 0], [0, 1]], dtype=complex)
    matrix2 = np.array([[1, 0], [0, 1]], dtype=complex)
    assert __are_equal_up_to_global_phase(matrix1, matrix2)

    matrix1 = np.zeros((2, 2), dtype=complex)
    matrix2 = np.zeros((2, 2), dtype=complex)
    assert __are_equal_up_to_global_phase(matrix1, matrix2)

    matrix1 = np.array([[1e10, 0], [0, 1e10]], dtype=complex)
    matrix2 = np.exp(1j * np.pi / 3) * matrix1
    assert __are_equal_up_to_global_phase(matrix1, matrix2)

    matrix1 = np.array([[np.nan, 0], [0, 1]], dtype=complex)
    matrix2 = np.array([[1, 0], [0, 1]], dtype=complex)
    assert not __are_equal_up_to_global_phase(matrix1, matrix2)


def test_rx():
    theta = 0
    expected = np.array([[1, 0], [0, 1]], dtype=complex)
    assert np.allclose(rx(theta), expected)

    theta = np.pi
    expected = np.array([[0, -1j], [-1j, 0]], dtype=complex)
    assert np.allclose(rx(theta), expected)

    theta = -np.pi / 2
    expected = np.array([[np.sqrt(2) / 2, 1j * np.sqrt(2) / 2], [1j * np.sqrt(2) / 2, np.sqrt(2) / 2]],
                        dtype=complex)
    assert np.allclose(rx(theta), expected)

    rng = np.random.default_rng(123)
    theta = rng.uniform(-2 * np.pi, 2 * np.pi)
    matrix = rx(theta)
    identity = np.eye(2, dtype=complex)
    assert np.allclose(matrix @ matrix.conj().T, identity)


def test_rz():
    theta = 0
    expected = np.array([[1, 0], [0, 1]], dtype=complex)
    assert np.allclose(rz(theta), expected)

    theta = np.pi
    expected = np.array([[np.exp(-1j * np.pi / 2), 0], [0, np.exp(1j * np.pi / 2)]], dtype=complex)
    assert np.allclose(rz(theta), expected)

    theta = -np.pi / 2
    expected = np.array([[np.exp(1j * np.pi / 4), 0], [0, np.exp(-1j * np.pi / 4)]], dtype=complex)
    assert np.allclose(rz(theta), expected)

    rng = np.random.default_rng(123)
    theta = rng.uniform(-2 * np.pi, 2 * np.pi)
    matrix = rz(theta)
    identity = np.eye(2, dtype=complex)
    assert np.allclose(matrix @ matrix.conj().T, identity)


def test___generate_two_qubit_gates():
    gates_dict = {'I': np.eye(2, dtype=complex), 'X': np.array([[0, 1], [1, 0]], dtype=complex)}
    result = __generate_two_qubit_gates(gates_dict)
    expected_keys = {'cz', 'I⊗X', 'cx', 'X⊗X', 'swap', 'iswap', 'X⊗I', 'I⊗I'}
    assert set(result.keys()) == expected_keys

    gates_dict = {}
    result = __generate_two_qubit_gates(gates_dict)
    expected_result = {
        'cx': np.array([[1, 0, 0, 0],
                        [0, 0, 0, 1],
                        [0, 0, 1, 0],
                        [0, 1, 0, 0]], dtype=complex),
        iSwapGate().name: iSwapGate().to_matrix(),
        CZGate().name: CZGate().to_matrix(),
        SwapGate().name: SwapGate().to_matrix()
    }
    assert set(result.keys()) == set(expected_result.keys())
    assert np.allclose(result['cx'], expected_result['cx'])
    assert np.allclose(result['iswap'], expected_result['iswap'])

    custom_gate = np.array([[1, 1], [1, -1]], dtype=complex)
    gates_dict = {'Custom': custom_gate}
    result = __generate_two_qubit_gates(gates_dict)
    expected_key = 'Custom⊗Custom'
    assert expected_key in result
    assert np.allclose(result[expected_key], np.kron(custom_gate, custom_gate))


def test_generate_two_qubit_gates():
    result = generate_two_qubit_clifford_gates()
    expected_keys = {'id⊗id', 'x⊗x', 'y⊗y', 'z⊗z', 'h⊗h', 's⊗s'}
    assert set(result.keys()).issuperset(expected_keys)

    result = generate_two_qubit_clifford_gates()
    expected_keys = {'Rx(pi/2)⊗Rx(pi/2)', 'Rz(pi/2)⊗Rz(pi/2)', 'Rx(pi)⊗Rx(pi)', 'Rz(pi)⊗Rz(pi)'}
    assert set(result.keys()).issuperset(expected_keys)

    result = generate_two_qubit_clifford_gates()
    expected_keys = {'cx', 'iswap', 'cz', 'swap'}
    assert set(result.keys()).issuperset(expected_keys)

    gates_dict = {}
    result = __generate_two_qubit_gates(gates_dict)
    expected_keys = {'cx', 'iswap', 'cz', 'swap'}
    assert set(result.keys()) == expected_keys

    custom_gate = np.array([[1, 1], [1, -1]], dtype=complex)
    gates_dict = {'Custom': custom_gate}
    result = __generate_two_qubit_gates(gates_dict)
    expected_key = 'Custom⊗Custom'
    assert expected_key in result
    assert np.allclose(result[expected_key], np.kron(custom_gate, custom_gate))


def test___multiply_sequence():
    matrices = [np.array([[2, 0], [0, 2]], dtype=np.complex128)]
    result = __multiply_sequence(matrices)
    expected = matrices[0]
    assert np.allclose(result, expected)

    matrices = [
        np.array([[1, 2], [3, 4]], dtype=np.complex128),
        np.array([[0, 1], [1, 0]], dtype=np.complex128),
        np.array([[2, 0], [0, 2]], dtype=np.complex128)
    ]
    result = __multiply_sequence(matrices)
    expected = matrices[0] @ matrices[1] @ matrices[2]
    assert np.allclose(result, expected)

    matrices = [
        np.array([[1, 2, 3], [4, 5, 6]], dtype=np.complex128),
        np.array([[7, 8], [9, 10], [11, 12]], dtype=np.complex128)
    ]
    result = __multiply_sequence(matrices)
    expected = matrices[0] @ matrices[1]
    assert np.allclose(result, expected)

    matrices = [
        np.array([[1, 2], [3, 4]], dtype=np.complex128),
        np.array([[1, 2, 3]], dtype=np.complex128)
    ]
    try:
        __multiply_sequence(matrices)
        assert False, "Expected ValueError for incompatible matrices"
    except ValueError:
        pass


def test_is_clifford_gate():
    assert is_clifford_gate(HGate())
    assert not is_clifford_gate(TGate())


def test___try_instantiate_gate():
    gates = __try_instantiate_gate(XGate)
    assert len(gates) == 1
    assert isinstance(gates[0], Instruction)
    assert isinstance(gates[0], XGate)

    gates = __try_instantiate_gate(RZGate)
    assert len(gates) == 13
    assert isinstance(gates[0], Instruction)
    assert isinstance(gates[0], RZGate)

    gates = __try_instantiate_gate(np.complex128)
    assert len(gates) == 0

    gates = __try_instantiate_gate(CXGate)
    assert len(gates) == 1
    assert isinstance(gates[0], Instruction)
    assert isinstance(gates[0], CXGate)

    gates = __try_instantiate_gate(MCXGate, num_control_qubits=3)
    assert len(gates) == 1
    assert isinstance(gates[0], Instruction)
    assert isinstance(gates[0], MCXGate)

    gates = __try_instantiate_gate(MCXGate)
    assert len(gates) == 1
    assert isinstance(gates[0], Instruction)
    assert isinstance(gates[0], CXGate)  # Should resolve to 1-control CX gate


def test_get_gates():
    gates = get_gates()
    assert "XGate" in gates
    assert "HGate" in gates
    assert "RZGate(1pi/4)" in gates
    assert isinstance(gates["XGate"], XGate)
    assert isinstance(gates["RZGate(1pi/4)"], RZGate)


def test_get_non_clifford_gates():
    gates = get_non_clifford_gates()
    assert len(gates) > 0
    assert "TGate" in gates
    assert isinstance(gates["TGate"], TGate)
    assert "HGate" not in gates



from _pytest.python_api import raises
from qiskit import QuantumCircuit
from qiskit.quantum_info import Operator, Statevector
from qiskit.transpiler import CouplingMap

from stabiliser_suite import circuit


def test_cnot():
    # Verify that CNOT is applied correctly when no coupling map is given
    qc = QuantumCircuit(3)
    circuit.cnot(qc, 0, 2)
    assert len(qc.data) == 1
    assert qc.data[0].name == 'cx'
    assert qc.data[0].qubits[0]._index == 0
    assert qc.data[0].qubits[1]._index == 2

    # Verify that error is raised in case of impossible operation
    coupling_map = CouplingMap([[0, 1], [2, 3]])
    qc = QuantumCircuit(4)
    with raises(ValueError):
        circuit.cnot(qc, 0, 3, coupling_map=coupling_map)

    # Verify that CNOT is applied correctly with a coupling map with direct connection
    coupling_map = CouplingMap([[0, 1], [1, 2]])
    qc = QuantumCircuit(3)
    circuit.cnot(qc, 1, 2, coupling_map=coupling_map)
    assert len(qc.data) == 1
    assert qc.data[0].name == 'cx'
    assert qc.data[0].qubits[0]._index == 1
    assert qc.data[0].qubits[1]._index == 2

    # Verify that CNOT is applied correctly with a coupling map with indirect connection
    coupling_map = CouplingMap([[0, 1], [2, 3], [1, 2]])
    qc = QuantumCircuit(4)
    circuit.cnot(qc, 0, 2, coupling_map=coupling_map)
    assert len(qc.data) == 4
    assert qc.data[0].name == 'cx'
    assert qc.data[0].qubits[0]._index == 0
    assert qc.data[0].qubits[1]._index == 1
    assert qc.data[1].name == 'cx'
    assert qc.data[1].qubits[0]._index == 1
    assert qc.data[1].qubits[1]._index == 2
    assert qc.data[2].name == 'cx'
    assert qc.data[2].qubits[0]._index == 0
    assert qc.data[2].qubits[1]._index == 1
    assert qc.data[3].name == 'cx'
    assert qc.data[3].qubits[0]._index == 1
    assert qc.data[3].qubits[1]._index == 2

    # Verify that the resulting operation is the same
    coupling_map = CouplingMap([[0, 1], [2, 3], [1, 2]])
    qc_test = QuantumCircuit(4)
    qc_control = QuantumCircuit(4)

    circuit.cnot(qc_test, 0, 2, coupling_map=coupling_map)
    qc_control.cx(0, 2)
    u_test = Operator(qc_test)
    u_control = Operator(qc_control)
    assert u_control.equiv(u_test)



"""
Benchmark for find_non_clifford_substitutions
=============================================
Save as benchmark_transformation.py and run with
$ python benchmark_transformation.py
Requires: matplotlib, pandas, qiskit, and your toolbox in PYTHONPATH.
"""
import time
import itertools
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from qiskit.circuit.library import HGate, CXGate

from stabiliser_suite.gate_utils import find_non_clifford_transformations

targets = [("CX (two qubit)", CXGate())]
depths  = list(range(1, 11))          # test max_depth = 1 … 9
rows    = []

for name, gate in targets:
    for d in depths:
        t0 = time.perf_counter()
        subs = find_non_clifford_transformations(gate, max_depth=d)
        elapsed = time.perf_counter() - t0
        rows.append({
            "gate": name,
            "max_depth": d,
            "runtime_s": elapsed,
            "solutions": len(subs),
        })
        print(f"{name:18}  depth={d}  {elapsed:7.3f}s  solutions={len(subs)}")

df = pd.DataFrame(rows)
print("\n=== Raw results ===")
print(df)

# ------------------------------------------------------------
# Plot runtime
for name in df["gate"].unique():
    sub = df[df["gate"] == name]
    plt.figure()
    plt.plot(sub["max_depth"], sub["runtime_s"], marker="o")
    plt.title(f"Runtime vs depth — {name}")
    plt.xlabel("max_depth")
    plt.ylabel("seconds")
    plt.grid(True)
    plt.tight_layout()

# Plot solutions
for name in df["gate"].unique():
    sub = df[df["gate"] == name]
    plt.figure()
    plt.plot(sub["max_depth"], sub["solutions"], marker="s")
    plt.title(f"Solutions vs depth — {name}")
    plt.xlabel("max_depth")
    plt.ylabel("# solutions")
    plt.grid(True)
    plt.tight_layout()

plt.show()


from qiskit import QuantumCircuit
from qiskit.transpiler import CouplingMap, CouplingError


def cnot(qc: QuantumCircuit, control: int, target: int, coupling_map: CouplingMap = None) -> None:
    """
    Applies a CNOT gate to the given quantum circuit.
    Attempts to find a bridge between the control and target qubits if the qubits cannot interact directly.

    :param qc: Circuit to apply the CNOT gate to
    :param control: Control qubit
    :param target: Target qubit
    :param coupling_map: Coupling map of the quantum circuit
    :return: None
    """
    # If no coupling map is provided, apply the CNOT gate directly
    if coupling_map is None:
        qc.cx(control, target)
        return

    # Check if the control and target qubits can be connected
    try:
        distance = coupling_map.distance(control, target)
    except CouplingError:
        raise ValueError('Control and target qubits are not connected in the coupling map.')

    # If the qubits are directly connected, apply the CNOT gate
    if distance == 1:
        qc.cx(control, target)
        return

    # If the qubits are not directly connected, find a bridge
    path = coupling_map.shortest_undirected_path(control, target)

    # Apply the CNOT gate using the bridge to every path triplet
    for i in range(len(path) - 2):
        intermediate_control = path[i]
        bridge = path[i + 1]
        intermediate_target = path[i + 2]

        # Apply a CNOT gate between the intermediate control and intermediate target using the bridge
        qc.cx(intermediate_control, bridge)
        qc.cx(bridge, intermediate_target)
        qc.cx(intermediate_control, bridge)
        qc.cx(bridge, intermediate_target)



import numpy as np


def __find_qubit_count(tableau: np.ndarray) -> int:
    """
    Finds the amount of qubits present in the tableau.

    :param tableau: Tableau representation of the circuit
    :return: Amount of qubits present in the tableau
    """
    return (tableau.shape[1] - 1) // 2


def __do_paulis_anticommute(tableau: np.ndarray) -> bool:
    qubit_count = __find_qubit_count(tableau)

    x1_z2_product = 0
    x2_z1_product = 0
    for i in range(qubit_count):
        x1_z2_product += (tableau[0][i] + tableau[1][i + qubit_count])
        x2_z1_product += (tableau[1][i] + tableau[0][i + qubit_count])

    return False if (x1_z2_product + x2_z1_product) % 2 == 0 else True



import numpy as np
import qiskit
from qiskit import QuantumCircuit
from qiskit.transpiler import CouplingMap, CouplingError
import termtables
from stabiliser_suite import utils


def sample_clifford_group(qubit_count: int, add_barriers: bool = False, coupling_map: CouplingMap = None,
                          max_distance: int = 1) -> QuantumCircuit:
    """
    Samples random Clifford operator.
    Uses algorithm presented by Ewout van den Berg (https://arxiv.org/pdf/2008.06011).
    :param qubit_count: The amount of qubits in the generated circuit.
    :param add_barriers: Whether the barriers should be added. NB! Barriers might slow down execution.
    :param coupling_map: The coupling map of the generated circuit.
    :param max_distance: Maximum allowed distance between qubits in the coupling map.
    :return: Generated quantum circuit.
    """
    qc: QuantumCircuit = qiskit.QuantumCircuit(qubit_count, qubit_count)

    for i in range(qubit_count):
        # Generate rows until the anticommuting Paulis are found
        row: np.ndarray = __generate_anticommuting_tableau(qubit_count - i)

        # Perform sweeping
        step_circuit: QuantumCircuit = qiskit.QuantumCircuit(qubit_count, qubit_count)
        while not __perform_sweeping(row, step_circuit, iteration=i, coupling_map=coupling_map,
                                     max_distance=max_distance):
            step_circuit.clear()
            row = __generate_anticommuting_tableau(qubit_count - i)

        qc.compose(step_circuit, inplace=True)

        if add_barriers:
            qc.barrier()

    return qc


def __perform_sweeping(tableau: np.ndarray, qc: QuantumCircuit, iteration: int = 0,
                       coupling_map: CouplingMap = None, max_distance: int = 1) -> bool:
    """
    Performs sweeping operation on the tableau.
    :param tableau: The tableau on which the sweeping operation is performed.
    :param qc: Quantum circuit which is being modified.
    :param iteration: Which sweeping operation is currently being performed.
    :param coupling_map: Quantum computer qubit coupling map.
    :return: True if tableau could be normalised successfully, False otherwise.
    """
    # Normalisation has to be checked before each step in order to avoid additional computations
    if __check_normalisation(tableau):
        return True
    __step_1(tableau, qc, shift=iteration)
    if __check_normalisation(tableau):
        return True
    if not __step_2(tableau, qc, shift=iteration, coupling_map=coupling_map, max_distance=max_distance):
        return False
    if __check_normalisation(tableau):
        return True
    if not __step_3(tableau, qc, shift=iteration, coupling_map=coupling_map, max_distance=max_distance):
        return False
    if __check_normalisation(tableau):
        return True
    __step_4(tableau, qc, shift=iteration)
    if __check_normalisation(tableau):
        return True

    # Apply steps 1 and 2 to the second row
    __step_1(tableau, qc, row_index=1, shift=iteration)
    if __check_normalisation(tableau):
        return True
    if not __step_2(tableau, qc, row_index=1, shift=iteration, coupling_map=coupling_map, max_distance=max_distance):
        return False
    if __check_normalisation(tableau):
        return True

    # Repeat step 4
    __step_4(tableau, qc, shift=iteration)
    if __check_normalisation(tableau):
        return True
    __step_5(tableau, qc, shift=iteration)

    return True


def __step_1(tableau: np.ndarray, qc: QuantumCircuit, row_index: int = 0, shift: int = 0) -> None:
    """
    Performs the first step of the sweeping algorithm by clearing the Z-block of the given row.
    Column indexing is shifted from actual by shift value.
    :param tableau: Tableau on which the operation is being performed.
    :param qc: Quantum circuit which is being modified.
    :param row_index: Index of the row of the tableau.
    :param shift: Column index shift, corresponds to the sweeping iteration.
    :return: None.
    """
    starting_index: int = utils.__find_qubit_count(tableau)
    for i in range(starting_index):
        if tableau[row_index][starting_index + i]:
            if tableau[row_index][i]:
                __apply_s(tableau, qc, i, shift)
            else:
                __apply_hadamard(tableau, qc, i, shift)


def __step_2(tableau: np.ndarray,
             qc: QuantumCircuit,
             row_index: int = 0,
             shift: int = 0, coupling_map: CouplingMap = None, max_distance: int = 1) -> bool:
    """
    Performs the second step of the sweeping algorithm by leaving only one operation in X-block.
    Column indexing is shifted from actual by shift value.
    :param tableau: Tableau on which the operation is being performed.
    :param qc: Quantum circuit which is being modified.
    :param row_index: Index of the row of the tableau.
    :param shift: Column index shift, corresponds to the sweeping iteration.
    :param coupling_map: Quantum computer qubit coupling map.
    :return: True if step can be completed with provided coupling map.
    """
    # Determine the set of indices where the coefficient is non-zero
    coefficients = [
        i for i in range(utils.__find_qubit_count(tableau)) if tableau[row_index][i]
    ]

    while len(coefficients) > 1:
        # Apply CNOT gate pairwise to qubits at even locations
        for i in range(len(coefficients) - 1):
            if i % 2 == 0:
                # Verify whether CNOT can be applied
                if coupling_map is not None:
                    try:
                        if coupling_map.distance(coefficients[i] + shift, coefficients[i + 1] + shift) > max_distance:
                            return False
                    except CouplingError:
                        return False
                __apply_cnot(tableau, qc, coefficients[i], coefficients[i + 1], shift=shift)

        # Retain indices at even locations
        coefficients = coefficients[::2]

    return True


def __step_3(tableau: np.ndarray, qc: QuantumCircuit, shift: int = 0, coupling_map: CouplingMap = None,
             max_distance: int = 1) -> bool:
    """
    Performs the third step of the sweeping algorithm by putting non-zero X-element into the first column.
    Column indexing is shifted from actual by shift value.
    :param tableau: Tableau on which the operation is being performed.
    :param qc: Quantum circuit which is being modified.
    :param shift: Column index shift, corresponds to the sweeping iteration.
    :param coupling_map: Quantum computer qubit coupling map.
    :return: True if step can be completed with provided coupling map.
    """
    # Determine the set of indices where the coefficient is non-zero
    coefficients: list[int] = list()
    for i in range((tableau.shape[1] - 1) // 2):
        if tableau[0][i]:
            coefficients.append(i)

    if len(coefficients) > 1:
        raise RuntimeError(f"Too many coefficients after step 2: {coefficients}")

    if len(coefficients) == 1 and coefficients[0] != 0:
        if coupling_map is not None:
            try:
                if coupling_map.distance(0 + shift, coefficients[0] + shift) > max_distance:
                    return False
            except CouplingError:
                return False
        __apply_swap(tableau, qc, 0, coefficients[0], shift)

    return True


def __step_4(tableau: np.ndarray, qc: QuantumCircuit, shift: int = 0) -> None:
    """
    Performs the fourth step of the sweeping algorithm by clearing the Z-block of the second row.
    Column indexing is shifted from actual by shift value.
    :param tableau: Tableau on which the operation is being performed.
    :param qc: Quantum circuit which is being modified.
    :param shift: Column index shift, corresponds to the sweeping iteration.
    :return: None.
    """
    z_start: int = utils.__find_qubit_count(tableau)
    # If the second Pauli is equal to +- Z_1, we skip this step
    if not tableau[1][0] and tableau[1][z_start]:
        return

    __apply_hadamard(tableau, qc, 0, shift=shift)


def __step_5(tableau: np.ndarray, qc: QuantumCircuit, shift: int = 0) -> None:
    """
    Performs the fifth step of the sweeping algorithm by clearing sign bits.
    Column indexing is shifted from actual by shift value.
    :param tableau: Tableau on which the operation is being performed.
    :param qc: Quantum circuit which is being modified.
    :param shift: Column index shift, corresponds to the sweeping iteration.
    :return: None.
    """
    phase_index: int = tableau.shape[1] - 1
    if not tableau[0][phase_index] and tableau[1][phase_index]:
        __apply_pauli(tableau, qc, 0, 'x', shift=shift)
    if tableau[0][phase_index]:
        if tableau[1][phase_index]:
            __apply_pauli(tableau, qc, 0, 'y', shift=shift)
        else:
            __apply_pauli(tableau, qc, 0, 'z', shift=shift)


def __generate_random_tableau(qubit_count: int) -> np.ndarray:
    """
    Generates a random tableau row.

    :param qubit_count: The amount of qubits in a tableau.
    :return: Randomly sampled tableau.
    """
    rng = np.random.default_rng()
    return rng.choice([False, True], size=(2, 2 * qubit_count + 1))


def __generate_anticommuting_tableau(qubit_count: int) -> np.ndarray:
    """
    Generates random tableau row with anticommuting Paulis.
    :param qubit_count: The amount of qubits in a tableau.
    :return: Randomly sampled tableau with anticommuting Paulis.
    """
    row = __generate_random_tableau(qubit_count)
    while not __check_if_tableau_suits(row):
        row = __generate_random_tableau(qubit_count)

    return row


def __apply_hadamard(tableau: np.ndarray, qc: QuantumCircuit, qubit: int, shift: int = 0) -> None:
    """
    Applies Hadamard gate to the circuit and modifies tableau.
    :param tableau: Tableau to be modified.
    :param qc: Quantum circuit which is being modified.
    :param qubit: The qubit to which the gate is applied.
    :param shift: Shift in qubit indexing.
    :return: None.
    """
    qc.h(qubit + shift)
    z_start: int = utils.__find_qubit_count(tableau)

    # Swaps columns in the tableau representation
    for row in tableau:
        temp = row[qubit]
        row[qubit] = row[z_start + qubit]
        row[z_start + qubit] = temp


def __apply_s(tableau: np.ndarray, qc: QuantumCircuit, qubit: int, shift: int = 0) -> None:
    """
    Applies S (phase shift) gate to the circuit and modifies tableau.
    :param tableau: Tableau to be modified.
    :param qc: Quantum circuit which is being modified.
    :param qubit: The qubit to which the gate is applied.
    :param shift: Shift in qubit indexing.
    :return: None.
    """
    qc.s(qubit + shift)
    z_start: int = utils.__find_qubit_count(tableau)

    # Swaps columns in the tableau representation
    for row in tableau:
        row[z_start + qubit] = row[z_start + qubit] != row[qubit]


def __apply_cnot(tableau: np.ndarray, qc: QuantumCircuit, control_qubit: int, target_qubit: int,
                 shift: int = 0) -> None:
    """
    Applies Hadamard gate to the circuit and modifies tableau.
    :param tableau: Tableau to be modified.
    :param qc: Quantum circuit which is being modified.
    :param control_qubit: The control qubit to which the gate is applied.
    :param target_qubit: The target qubit to which the gate is applied.
    :param shift: Shift in qubit indexing.
    :return: None.
    """
    qc.cx(control_qubit + shift, target_qubit + shift)
    z_start: int = utils.__find_qubit_count(tableau)

    # Swaps columns in the tableau representation
    for row in tableau:
        row[target_qubit] = row[target_qubit] != row[control_qubit]
        row[z_start + control_qubit] = row[z_start + control_qubit] != row[z_start + target_qubit]


def __apply_swap(tableau: np.ndarray, qc: QuantumCircuit, qubit_1: int, qubit_2: int, shift: int = 0) -> None:
    """
    Applies Swap (three cnot) gate to the circuit and modifies tableau.
    :param tableau: Tableau to be modified.
    :param qc: Quantum circuit which is being modified.
    :param qubit_1: The first qubit to which the gate is applied.
    :param qubit_2: The second qubit to which the gate is applied.
    :param shift: Shift in qubit indexing.
    :return: None.
    """
    qc.swap(qubit_1 + shift, qubit_2 + shift)  # Shift already applied to qubit 2
    z_start: int = utils.__find_qubit_count(tableau)
    for row in tableau:
        # X-swap
        temp = row[qubit_1]
        row[qubit_1] = row[qubit_2]
        row[qubit_2] = temp

        # Z-swap
        temp = row[qubit_1 + z_start]
        row[qubit_1 + z_start] = row[qubit_2 + z_start]
        row[qubit_2 + z_start] = temp


def __check_normalisation(tableau: np.ndarray) -> bool:
    """
    Verifies tableau normalisation.
    :param tableau: Tableau to be checked.
    :return: True if tableau is normalised.
    """
    z_start: int = utils.__find_qubit_count(tableau)
    return np.array_equal(np.nonzero(tableau[0])[0], [0]) and np.array_equal(np.nonzero(tableau[1])[0], [z_start])


def __get_pauli(tableau: np.ndarray, row: int, qubit: int) -> str:
    """
    Calculates the Pauli gate coded by the given qubit in the tableau.
    :param tableau: Tableau representation the circuit.
    :param row: Row for which the Pauli gate is calculated.
    :param qubit: Index of qubit.
    :return: Name of the Pauli gate.
    """
    z_start: int = utils.__find_qubit_count(tableau)
    if tableau[row][qubit]:
        if tableau[row][z_start + qubit]:
            return 'y'
        else:
            return 'x'
    else:
        if tableau[row][z_start + qubit]:
            return 'z'

    return 'I'


def __apply_pauli(tableau: np.ndarray, qc: QuantumCircuit, qubit: int, pauli: str, shift: int = 0) -> None:
    """
    Applies Pauli gate to the circuit and modifies tableau.
    :param tableau: Tableau to be modified.
    :param qc: Quantum circuit which is being modified.
    :param qubit: The qubit to which the gate is applied.
    :param pauli: Pauli gate.
    :param shift: Shift in qubit indexing.
    :return: None.
    """
    pauli_lower = pauli.lower()
    if pauli_lower == 'x':
        qc.x(qubit + shift)
    elif pauli_lower == 'y':
        qc.y(qubit + shift)
    elif pauli_lower == 'z':
        qc.z(qubit + shift)
    else:
        raise ValueError(f"Unknown pauli: {pauli}")
    for i, row in enumerate(tableau):
        row_pauli = __get_pauli(tableau, i, qubit)
        if row_pauli != pauli_lower and row_pauli != 'I':
            row[-1] = not row[-1]


def __check_if_tableau_suits(tableau: np.ndarray) -> bool:
    # Verify that the first row is not identity
    first_pauli = __get_pauli(tableau, 0, 0)
    if first_pauli == 'I':
        return False

    # Calculate the sympathetic product
    return utils.__do_paulis_anticommute(tableau)


def print_tableau(tableau: np.ndarray, representation: str = "text") -> None:
    """
    Prints tableau representation of the circuit.
    :param tableau: Tableau to be printed.
    :param representation: The representation model. "text" for text, and "mpl" for matplotlib.
    :return: None.
    """
    qubit_count: int = utils.__find_qubit_count(tableau)

    if representation == "text":
        header: list[str] = list()
        for i in range(qubit_count):
            header.append(f"X{i}")
        for i in range(qubit_count):
            header.append(f"Z{i}")
        header.append("S")

        termtables.print(np.vectorize(lambda t: "1" if t else "0")(tableau), header)
        return
    if representation == "mpl":
        raise NotImplementedError("Matplotlib representation is not implemented. See issue #8")
    else:
        raise ValueError(f"Unknown representation: {representation}")
